from xmlsig.utils import create_node

from .ns import EtsiNS


class ObjectIdentifier(object):
    def __init__(self, identifier, description=None, references=None):
        self.identifier = identifier
        self.description = description
        self.references = references or []

    def to_xml(self, node):
        create_node("Identifier", node, EtsiNS).text = self.identifier
        if self.description is not None:
            create_node("Description", node, EtsiNS).text = self.description
        if len(self.references) > 0:
            documentation = create_node("DocumentationReferences", node, EtsiNS)
            for reference in self.references:
                create_node(
                    "DocumentationReference", documentation, EtsiNS
                ).text = reference



# © 2017 Creu Blanca
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from uuid import uuid4

from .constants import ID_ATTR


def dict_compare(d1, d2):
    assert len(d1) == len(d2)
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    assert len(intersect_keys) == len(d1)
    for key in d1_keys:
        assert d1[key] == d2[key]


def rdns_to_map(data):
    return {
        x.split("=")[0].strip(): x.split("=")[1].strip() for x in data.split(",") if x
    }


def get_unique_id():
    return "id-{}".format(uuid4())


def ensure_id(node):
    """Ensure given node has a wsu:Id attribute; add unique one if not.

    Return found/created attribute value.

    """
    assert node is not None
    id_val = node.get(ID_ATTR)
    if not id_val:
        id_val = get_unique_id()
        node.set(ID_ATTR, id_val)
    return id_val



# © 2017 Creu Blanca
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

import hashlib
import logging
from base64 import b64decode, b64encode

from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_der_x509_certificate
from lxml.builder import ElementMaker
from xades.constants import MAP_HASHLIB, NS_MAP
from xades.ns import EtsiNS
from xades.utils import dict_compare, rdns_to_map
from xmlsig.constants import TransformSha1, TransformUsageDigestMethod
from xmlsig.ns import DSigNs
from xmlsig.utils import USING_PYTHON2, create_node, get_rdns_name

if USING_PYTHON2:
    import urllib
else:
    import urllib.request as urllib

logger = logging.getLogger(__name__)


ETSI = ElementMaker(namespace=EtsiNS)
DS = ElementMaker(namespace=DSigNs)


class BasePolicy(object):
    """ "
    Policy base class created in order to define different policies.
    A mixture of base class implementations, and abstract class
    interface definitions. (TODO: might be separated in the future)
    """

    hash_method = None

    @property
    def identifier(self):
        raise NotImplementedError("Implement on specific subclasses")

    @property
    def name(self):
        raise NotImplementedError("Implement on specific subclasses")

    @property
    def policy(self):
        raise NotImplementedError("Implement on specific subclasses")

    def _resolve_policy(self, identifier):
        """
        Resolves the signature policy to bytes.
        Override for resolving e.g. from a local cache
        :param identifier: the value of <etsi:SigPolicyId/etsi:Identifier/>
        :return: bytes
        """
        return urllib.urlopen(identifier).read()

    def sign(self, signature):
        return

    def validate(self, signature):
        """
        Finds if the policy is the same and then applies the policy validation.
        Otherwise, it does nothing
        :param signature: Signature node
        :return:
        """
        policy = signature.find(
            "ds:Object/etsi:QualifyingProperties/etsi:SignedProperties/"
            "etsi:SignedSignatureProperties/etsi:SignaturePolicyIdentifier/"
            "etsi:SignaturePolicyId",
            namespaces=NS_MAP,
        )
        if policy is None:
            return
        if self.identifier != policy.find(
            "etsi:SigPolicyId/etsi:Identifier", namespaces=NS_MAP
        ):
            return
        self.validate_policy(signature)

    def validate_policy(self, signature):
        """
        Policy validation
        :param signature: signature node
        :return: None
        """
        return

    def set_transforms(self, transforms, value, sign=False):
        """
        Creates transformations of the policy if required. Modifies node and
        returns the transformed value
        :param node: Policy node
        :param value: Original value
        :param sign: Calculates or validates the transformation
        :return: str
        """
        return value

    def _query_signature_policy_identifer_data(self, node):
        """
        Query common policy validation data.
        """
        signature_policy_id = node.find("etsi:SignaturePolicyId", namespaces=NS_MAP)
        sig_policy_id = signature_policy_id.find("etsi:SigPolicyId", namespaces=NS_MAP)
        identifier = sig_policy_id.find("etsi:Identifier", namespaces=NS_MAP).text
        hash_method = signature_policy_id.find(
            "etsi:SigPolicyHash/ds:DigestMethod", namespaces=NS_MAP
        ).get("Algorithm")
        digest_value = signature_policy_id.find(
            "etsi:SigPolicyHash/ds:DigestValue", namespaces=NS_MAP
        ).text
        transforms = signature_policy_id.find("ds:Tranforms", namespaces=NS_MAP)
        return {
            "Identifier": identifier,
            "DigestMethodAlgorithm": hash_method,
            "DigestValue": digest_value,
            "Transforms": transforms,
        }

    def validate_policy_node(self, node):
        """
        A validation implementation for a given
        <etsi:SignaturePolicyIdentifier/> node
        :param node: Policy node
        :return: bool
        """
        raise NotImplementedError("Implement on specific subclasses")

    def calculate_certificates(self, node, keys_x509):
        for key_x509 in keys_x509:
            self.calculate_certificate(node, key_x509)

    def calculate_certificate(self, node, key_x509):
        fingerprint = key_x509.fingerprint(MAP_HASHLIB[self.hash_method]())
        _ETSI_Cert = ETSI.Cert(
            ETSI.CertDigest(
                DS.DigestMethod(Algorithm=self.hash_method),
                DS.DigestValue(b64encode(fingerprint).decode()),
            ),
            ETSI.IssuerSerial(
                DS.X509IssuerName(get_rdns_name(key_x509.issuer.rdns)),
                DS.X509SerialNumber(str(key_x509.serial_number)),
            ),
        )
        node.append(_ETSI_Cert)

    def validate_certificate(self, node, signature):
        certs = node.findall("etsi:Cert", namespaces=NS_MAP)
        x509 = signature.find("ds:KeyInfo/ds:X509Data", namespaces=NS_MAP)
        x509_data = x509.find("ds:X509Certificate", namespaces=NS_MAP)
        serial = x509.find("ds:X509IssuerSerial", namespaces=NS_MAP)
        if serial is not None:
            serial_name = serial.find("ds:X509IssuerName", namespaces=NS_MAP).text
            serial_number = serial.find("ds:X509SerialNumber", namespaces=NS_MAP).text
            certificate = None
            for cert in certs:
                if (
                    cert.find(
                        "etsi:IssuerSerial/ds:X509IssuerName", namespaces=NS_MAP
                    ).text
                    == serial_name
                    and cert.find(
                        "etsi:IssuerSerial/ds:X509SerialNumber", namespaces=NS_MAP
                    ).text
                    == serial_number
                ):
                    certificate = cert
            assert certificate is not None
        else:
            certificate = certs[0]
        if x509_data is not None:
            parsed_x509 = load_der_x509_certificate(
                b64decode(x509_data.text), default_backend()
            )
            assert (
                str(parsed_x509.serial_number)
                == certificate.find(
                    "etsi:IssuerSerial/ds:X509SerialNumber", namespaces=NS_MAP
                ).text
            )
            dict_compare(
                rdns_to_map(get_rdns_name(parsed_x509.issuer.rdns)),
                rdns_to_map(
                    certificate.find(
                        "etsi:IssuerSerial/ds:X509IssuerName", namespaces=NS_MAP
                    ).text
                ),
            )
            digest = certificate.find("etsi:CertDigest", namespaces=NS_MAP)
            assert (
                b64encode(
                    parsed_x509.fingerprint(
                        MAP_HASHLIB[
                            digest.find("ds:DigestMethod", namespaces=NS_MAP).get(
                                "Algorithm"
                            )
                        ]()
                    )
                ).decode()
                == digest.find("ds:DigestValue", namespaces=NS_MAP).text
            )

    def calculate_policy_node(self, node, sign=False):
        """
        Calculates de policy node
        :param node: SignaturePolicyIdentifier node
        :param sign: checks if we must calculate or validate a policy
        :return:
        """
        logger.warning(
            "This method is deprecated. Use `produce_policy_node` "
            "or `validate_policy_node` accordingly."
        )
        if not sign:
            return self.validate_policy_node(node)
        return self.produce_policy_node(node)


class ImpliedPolicy(BasePolicy):
    def __init__(self, hash_method=TransformSha1):
        self.hash_method = hash_method

    @property
    def identifier(self):
        return None

    def produce_policy_node(self, node):
        """
        Produces the policy node
        :param node: SignaturePolicyIdentifier node
        :return:
        """
        create_node("SignaturePolicyImplied", node, EtsiNS)

    def validate_policy_node(self, node):
        """
        A specific validation implementation for a given
        <etsi:SignaturePolicyIdentifier/> node
        Implied policy by itself cannot be validated
        :param node: Policy node
        :return:
        """
        return


class GenericPolicyId(BasePolicy):
    def __init__(self, identifier, name, hash_method):
        self.generic_identifier = identifier
        self.generic_name = name
        self.hash_method = hash_method
        self._policy = None

    @property
    def identifier(self):
        return self.generic_identifier

    @property
    def name(self):
        return self.generic_name

    @property
    def policy(self):
        if not self._policy:
            self._policy = self._resolve_policy(self.identifier)
        return self._policy

    def produce_policy_node(self, node):
        """
        Produces the policy node
        :param node: SignaturePolicyIdentifier node
        :return:
        """
        value = self._resolve_policy(self.identifier)
        # Must be previously set in the template
        # TODO: Implement a better version
        transforms = node.find(
            "etsi:SignaturePolicyId/ds:Transforms", namespaces=NS_MAP
        )
        value = self.set_transforms(transforms, value, True)
        hash_calc = hashlib.new(TransformUsageDigestMethod[self.hash_method])
        hash_calc.update(value)
        _ETSI_SignaturePolicyId = ETSI.SignaturePolicyId(
            ETSI.SigPolicyId(
                ETSI.Identifier(self.identifier), ETSI.Description(self.name)
            ),
            ETSI.SigPolicyHash(
                DS.DigestMethod(Algorithm=self.hash_method),
                DS.DigestValue(b64encode(hash_calc.digest()).decode()),
            ),
        )
        node.append(_ETSI_SignaturePolicyId)

    def validate_policy_node(self, node):
        """
        A specifc validation implementation for a given
        <etsi:SignaturePolicyIdentifier/> node leveraging known cached policy
        :param node: Policy node
        :return: bool
        """
        implied = node.find("etsi:SignaturePolicyImplied", namespaces=NS_MAP)
        if implied is not None:
            return
        data = self._query_signature_policy_identifer_data(node)
        value = self.policy
        value = self.set_transforms(data["Transforms"], value, False)
        hash_calc = hashlib.new(
            TransformUsageDigestMethod[data["DigestMethodAlgorithm"]]
        )
        hash_calc.update(value)
        assert data["DigestValue"] == b64encode(hash_calc.digest()).decode()



# © 2017 Creu Blanca
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from cryptography.hazmat.primitives import hashes
from xmlsig import constants

from .ns import EtsiNS

ID_ATTR = "Id"
NS_MAP = constants.NS_MAP
NS_MAP["etsi"] = EtsiNS

MAP_HASHLIB = {
    constants.TransformMd5: hashes.MD5,
    constants.TransformSha1: hashes.SHA1,
    constants.TransformSha224: hashes.SHA224,
    constants.TransformSha256: hashes.SHA256,
    constants.TransformSha384: hashes.SHA384,
    constants.TransformSha512: hashes.SHA512,
}



from lxml import etree
from xmlsig.constants import DSigNs
from xmlsig.utils import create_node

from .constants import ID_ATTR, NS_MAP, EtsiNS


def create_qualifying_properties(node, name=None, etsi="etsi"):
    obj_node = create_node("Object", node, DSigNs)
    qualifying = etree.SubElement(
        obj_node, etree.QName(EtsiNS, "QualifyingProperties"), nsmap={etsi: EtsiNS}
    )
    qualifying.set("Target", "#" + node.get(ID_ATTR))
    if name is not None:
        qualifying.set(ID_ATTR, name)
    return qualifying


def create_signed_properties(node, name=None, datetime=None):
    properties = create_node("SignedProperties", node, EtsiNS)
    if name is not None:
        properties.set(ID_ATTR, name)
    signature_properties = create_node("SignedSignatureProperties", properties, EtsiNS)
    signing_time = create_node("SigningTime", signature_properties, EtsiNS)
    if datetime is not None:
        signing_time.text = datetime.isoformat()
    create_node("SigningCertificate", signature_properties, EtsiNS)
    create_node("SignaturePolicyIdentifier", signature_properties, EtsiNS)
    return properties


def add_production_place(node, city=None, state=None, postal_code=None, country=None):
    signature_properties = node.find(
        "etsi:SignedSignatureProperties", namespaces=NS_MAP
    )
    production_place = signature_properties.find(
        "etsi:SignatureProductionPlace", namespaces=NS_MAP
    )
    if production_place is None:
        production_place = create_node("SignatureProductionPlace", ns=EtsiNS)
        signature_properties.insert(3, production_place)
    for child in production_place.getchildren():
        production_place.remove(child)
    if city is not None:
        create_node("City", production_place, EtsiNS).text = city
    if state is not None:
        create_node("StateOrProvince", production_place, EtsiNS).text = state
    if postal_code is not None:
        create_node("PostalCode", production_place, EtsiNS).text = postal_code
    if country is not None:
        create_node("CountryName", production_place, EtsiNS).text = country


def add_claimed_role(node, role):
    signature_properties = node.find(
        "etsi:SignedSignatureProperties", namespaces=NS_MAP
    )
    signer_role = signature_properties.find("etsi:SignerRole", namespaces=NS_MAP)
    if signer_role is None:
        signer_role = create_node("SignerRole", signature_properties, ns=EtsiNS)
    claimed_roles = signer_role.find("etsi:ClaimedRoles", namespaces=NS_MAP)
    if claimed_roles is None:
        claimed_roles = create_node("ClaimedRoles", ns=EtsiNS)
        signer_role.insert(0, claimed_roles)
    claimed_role = create_node("ClaimedRole", claimed_roles, EtsiNS)
    claimed_role.text = role
    return claimed_role


def ensure_signed_data_object_properties(node):
    properties = node.find("etsi:SignedDataObjectProperties", namespaces=NS_MAP)
    if properties:
        return properties
    return create_node("SignedDataObjectProperties", node, EtsiNS)


def add_data_object_format(
    node, reference, description=None, identifier=None, mime_type=None, encoding=None
):
    data_object_node = create_node("DataObjectFormat", ns=EtsiNS)
    node.insert(
        len(node.findall("etsi:DataObjectFormat", namespaces=NS_MAP)), data_object_node
    )
    data_object_node.set("ObjectReference", reference)
    if description is not None:
        create_node("Description", data_object_node, EtsiNS).text = description
    if identifier is not None:
        identifier.to_xml(create_node("ObjectIdentifier", data_object_node, EtsiNS))
    if mime_type is not None:
        create_node("MimeType", data_object_node, EtsiNS).text = mime_type
    if encoding is not None:
        create_node("Encoding", data_object_node, EtsiNS).text = encoding
    return data_object_node


def add_commitment_type_indication(
    node, identifier, references=None, qualifiers_type=None
):
    commitment_type = create_node("CommitmentTypeIndication", ns=EtsiNS)
    node.insert(
        len(node.findall("etsi:DataObjectFormat", namespaces=NS_MAP))
        + len(node.findall("etsi:CommitmentTypeIndication", namespaces=NS_MAP)),
        commitment_type,
    )
    identifier.to_xml(create_node("CommitmentTypeId", commitment_type, EtsiNS))
    if references is None:
        create_node("AllSignedDataObjects", commitment_type, EtsiNS)
    else:
        for reference in references:
            create_node("ObjectReference", commitment_type, EtsiNS).text = reference
    if qualifiers_type is not None:
        qualifiers = create_node("CommitmentTypeQualifiers", commitment_type, EtsiNS)
        for qualifier in qualifiers_type:
            create_node("CommitmentTypeQualifier", qualifiers, EtsiNS).text = qualifier



# © 2017 Creu Blanca
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl.html).

from datetime import datetime
from os import path

import pytz
from lxml import etree
from xmlsig import SignatureContext

from .constants import NS_MAP
from .policy import ImpliedPolicy


class XAdESContext(SignatureContext):
    def __init__(self, policy=None, certificates=False):
        """
        Declaration
        :param policy: Policy class
        :type policy: xades.Policy
        :param certificates: List of certificates
        :type certificates: list<
            cryptography.hazmat.backends.openssl.x509._Certificate>
        """
        self.policy = policy
        self.policies = {None: ImpliedPolicy()}
        if policy is None:
            self.policy = ImpliedPolicy()
        else:
            self.policies[policy.identifier] = policy
        self.certificates = certificates
        super(XAdESContext, self).__init__()

    def sign(self, node):
        """
        Signs a node
        :param node: Signature node
        :type node: lxml.etree.Element
        :return: None
        """
        signed_properties = node.find(
            "ds:Object/etsi:QualifyingProperties["
            "@Target='#{}']/etsi:SignedProperties".format(node.get("Id")),
            namespaces=NS_MAP,
        )
        assert signed_properties is not None
        self.calculate_signed_properties(signed_properties, node, True)
        unsigned_properties = node.find(
            "ds:Object/etsi:QualifyingProperties["
            "@Target='#{}']/etsi:UnsignedProperties".format(node.get("Id")),
            namespaces=NS_MAP,
        )
        if unsigned_properties is not None:
            self.calculate_unsigned_properties(unsigned_properties, node, True)
        self.policy.sign(node)
        res = super(XAdESContext, self).sign(node)
        return res

    def verify(self, node):
        """
        verifies a signature
        :param node: Signature node
        :type node: lxml.etree.Element
        :return:
        """
        schema = etree.XMLSchema(
            etree.parse(path.join(path.dirname(__file__), "data/XAdES.xsd"))
        )
        schema.assertValid(node)
        signed_properties = node.find(
            "ds:Object/etsi:QualifyingProperties["
            "@Target='#{}']/etsi:SignedProperties".format(node.get("Id")),
            namespaces=NS_MAP,
        )
        assert signed_properties is not None
        self.calculate_signed_properties(signed_properties, node, False)
        unsigned_properties = node.find(
            "ds:Object/etsi:QualifyingProperties["
            "@Target='#{}']/etsi:UnSignedProperties".format(node.get("Id")),
            namespaces=NS_MAP,
        )
        if unsigned_properties is not None:
            self.calculate_unsigned_properties(signed_properties, node, False)
        policy_id = signed_properties.find(
            "etsi:SignedSignatureProperties/etsi:SignaturePolicyIdentifier/"
            "etsi:SignaturePolicyId/etsi:SigPolicyId/etsi:Identifier",
            namespaces=NS_MAP,
        )
        policy = self.policies[None]
        if policy_id is not None:
            if policy_id.text in self.policies:
                policy = self.policies[policy_id.text]
        policy.validate(node)
        res = super(XAdESContext, self).verify(node)
        return res

    def calculate_signed_properties(self, signed_properties, node, sign=False):
        signature_properties = signed_properties.find(
            "etsi:SignedSignatureProperties", namespaces=NS_MAP
        )
        assert signature_properties is not None
        self.calculate_signature_properties(signature_properties, node, sign)
        data_object_properties = signed_properties.find(
            "etsi:SignedDataObjectProperties", namespaces=NS_MAP
        )
        if data_object_properties is not None:
            self.calculate_data_object_properties(data_object_properties, node, sign)
        return

    def calculate_signature_properties(self, signature_properties, node, sign=False):
        signing_time = signature_properties.find("etsi:SigningTime", namespaces=NS_MAP)
        assert signing_time is not None
        if sign and signing_time.text is None:
            now = datetime.now().replace(microsecond=0, tzinfo=pytz.utc)
            signing_time.text = now.isoformat()
        certificate_list = signature_properties.find(
            "etsi:SigningCertificate", namespaces=NS_MAP
        )
        assert certificate_list is not None
        policy = signature_properties.find(
            "etsi:SignaturePolicyIdentifier", namespaces=NS_MAP
        )
        if sign:
            assert policy is not None
            self.policy.calculate_certificates(
                certificate_list, self.certificates or [self.x509]
            )
            self.policy.produce_policy_node(policy)
        else:
            self.policy.validate_certificate(certificate_list, node)
            self.policy.validate_policy_node(policy)

    def calculate_data_object_properties(
        self, data_object_properties, node, sign=False
    ):
        """
        To be improved with EPES, T...
        :param data_object_properties: DataObjectProperties node
        :param node: Signature node
        :param sign:
        :return:
        """
        return

    def calculate_unsigned_properties(self, unsigned_properties, node, sign=False):
        """
        To be improved with EPES, T...
        :param unsigned_properties: UnsignedProperties node
        :param node: Signature node
        :param sign:
        :return:
        """
        return



from setuptools import find_packages, setup

install_requires = [
    "xmlsig",
    "pytz",
]
test_requires = ["xmlsig[test]"]


setup(
    name="xades",
    version="1.0.0",
    description="XaDES XML signature",
    long_description="XaDES XML Signature created with cryptography and lxml",
    author="Enric Tobella Alomar",
    author_email="etobella@creublanca.es",
    url="http://github.com/etobella/python-xades",
    install_requires=install_requires,
    tests_require=test_requires,
    extras_require={"test": test_requires},
    entry_points={},
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    license="LGPL-3",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    zip_safe=False,
)
